/********************************************************************************
 * Copyright (c) 2023 CEA-List
 *
 * This program and the accompanying materials are made available under the
 * terms of the Eclipse Public License 2.0 which is available at
 * http://www.eclipse.org/legal/epl-2.0.
 *
 * SPDX-License-Identifier: EPL-2.0
 *
 ********************************************************************************/

#ifndef AIDGE_CORE_OPERATOR_BATCHNORM_H_
#define AIDGE_CORE_OPERATOR_BATCHNORM_H_

#include <array>
#include <memory>
#include <vector>

#include "aidge/utils/Types.h"
#include "aidge/data/Tensor.hpp"
#include "aidge/graph/Node.hpp"
#include "aidge/operator/OperatorTensor.hpp"
#include "aidge/operator/Producer.hpp"
#include "aidge/utils/StaticAttributes.hpp"
#include "aidge/utils/Registrar.hpp"

namespace Aidge {

enum class BatchNormAttr { Epsilon, Momentum };

template <DimIdx_t DIM>
class BatchNorm_Op : public OperatorTensor,
                public Registrable<BatchNorm_Op<DIM>, std::string, std::shared_ptr<OperatorImpl>(const BatchNorm_Op<DIM> &)>,
                public StaticAttributes<BatchNormAttr, float, float> {
public:
    static const std::string Type;

    BatchNorm_Op() = delete;

    using Attributes_ = StaticAttributes<BatchNormAttr, float, float>;
    template <BatchNormAttr e>
    using attr = typename Attributes_::template attr<e>;

    constexpr BatchNorm_Op(float epsilon, float momentum)
        : OperatorTensor(Type, 1, 4, 1),
          Attributes_(attr<BatchNormAttr::Epsilon>(epsilon),
                           attr<BatchNormAttr::Momentum>(momentum)) {}

    /**
     * @brief Copy-constructor. Copy the operator attributes and its output tensor(s), but not its input tensors (the new operator has no input associated).
     * @param op Operator to copy.
     */
    BatchNorm_Op(const BatchNorm_Op<DIM>& op)
        : OperatorTensor(op),
          Attributes_(op)
    {
        if (op.mImpl){
            SET_IMPL_MACRO(BatchNorm_Op<DIM>, *this, op.mOutputs[0]->getImpl()->backend());
        }else{
            mImpl = nullptr;
        }
    }

    /**
     * @brief Clone the operator using its copy-constructor.
     * @see Operator::BatchNorm_Op
     */
    std::shared_ptr<Operator> clone() const override {
        return std::make_shared<BatchNorm_Op<DIM>>(*this);
    }

    // Data operator[](const char* inputName) override final {
    //     std::shared_ptr<Tensor> in = (strcmp(inputName, "data")) ? mInputs[0] :
    //         (strcmp(inputName, "weight") ? mInputs[1] :
    //         (strcmp(inputName, "bias") ? mInputs[2] :
    //         nullptr));
    //     assert((in!=nullptr) && "No such parameter");
    //     return *in;
    // }


    void computeOutputDims() override final {
        // check inputs have been associated
        bool associated = true;
        for (IOIndex_t i = 0; i < nbInputs(); ++i) {
            associated &= !(getInput(i)->empty());
        }
        if (associated) {
            const DimSize_t nbFeatures =  getInput(0)->dims()[1];
            for (std::size_t i = nbData(); i < nbInputs(); ++i) {
                if(getInput(i)->size() != nbFeatures) {
                    // /!\ Input size should be handled BEFORE calling this function
                    // This should raise an error
                    getInput(i)->resize({getInput(0)->dims()[1]});
                }
            }
            mOutputs[0]->resize(getInput(0)->dims());
        }
    }

    void setBackend(const std::string &name, DeviceIdx_t device = 0) override {
        SET_IMPL_MACRO(BatchNorm_Op<DIM>, *this, name);
        mOutputs[0]->setBackend(name, device);

        // By default, automatically set backend for scale, shift, mean and variance
        getInput(1)->setBackend(name, device);
        getInput(2)->setBackend(name, device);
        getInput(3)->setBackend(name, device);
        getInput(4)->setBackend(name, device);
    }

    static const std::vector<std::string> getInputsName() {
        return {"data_input", "scale", "shift", "mean", "variance"};
    }
    static const std::vector<std::string> getOutputsName() {
        return {"data_output"};
    }
};

template <DimIdx_t DIM>
const std::string BatchNorm_Op<DIM>::Type = "BatchNorm";

template <DimSize_t DIM>
inline std::shared_ptr<Node> BatchNorm(const DimSize_t nbFeatures,
                                       const float epsilon = 1.0e-5F,
                                       const float momentum = 0.1F,
                                       const std::string& name = "") {
    static_assert(DIM<=MaxDim,"Too many kernel dimensions required by BatchNorm, not supported");
    auto batchNorm = std::make_shared<Node>(std::make_shared<BatchNorm_Op<static_cast<DimIdx_t>(DIM)>>(epsilon, momentum), name);
    addProducer(batchNorm, 1, {nbFeatures}, "scale");
    addProducer(batchNorm, 2, {nbFeatures}, "shift");
    addProducer(batchNorm, 3, {nbFeatures}, "batch_mean");
    addProducer(batchNorm, 4, {nbFeatures}, "batch_variance");
    return batchNorm;
}
}  // namespace Aidge

namespace {
template <>
const char *const EnumStrings<Aidge::BatchNormAttr>::data[] = { "Epsilon", "Momentum" };
}

#endif //AIDGE_CORE_OPERATOR_BATCHNORM_H_
